﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2019 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using HEAL.Attic;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Problems.DataAnalysis.Symbolic;
using System.Collections.Generic;
using System.IO;
using System.Linq;

namespace HeuristicLab.Algorithms.EvolvmentModelsOfModels {
  [StorableType("83CF9650-98FF-454B-9072-82EA4D39C752")]
  public abstract class EMMMapBase<T> : Item where T : class {
    [Storable]
    public List<T> ModelSet { get; set; }
    [Storable]
    public List<int> ClusterNumber { get; set; }
    [Storable]
    public List<List<int>> Map { get; set; }
    [Storable]
    public double[,] Distances { get; set; }
    [Storable]
    public int K { get; protected set; }

    protected void CalculateDistances() {
      if (ModelSet is List<ISymbolicExpressionTree> set) {
        Distances = SymbolicExpressionTreeHash.ComputeSimilarityMatrix(set, simplify: false, strict: true);
      } else { /// for future work
        for (int i = 0; i < ModelSet.Count - 1; i++) {
          for (int j = 0; j <= i; j++) {
            Distances[i, j] = 0;
          }
        }
      }
      for (int i = 0; i < ModelSet.Count - 1; i++) {
        for (int j = i + 1; j < ModelSet.Count; j++) {
          Distances[j, i] = Distances[i, j] = 1 - Distances[i, j];
        }
      }

    }
    public abstract void CreateMap(IRandom random, int k);
    public abstract T NewModelForInizializtionNotTree(IRandom random, out int cluster, out int treeNumber);
    public ISymbolicExpressionTree NewModelForInizializtion(IRandom random, out int cluster, out int treeNumber) {
      treeNumber = random.Next(ModelSet.Count);
      cluster = ClusterNumber[treeNumber];
      if (ModelSet[treeNumber] is ISymbolicExpressionTree model)
        return (ISymbolicExpressionTree)(model.Clone());
      return new SymbolicExpressionTree();
    }

    [StorableConstructor]
    protected EMMMapBase(StorableConstructorFlag _) : base(_) { }
    protected EMMMapBase() : this(1) { }
    public EMMMapBase(int k) {
      K = k;
      ClusterNumber = new List<int>();
      Map = new List<List<int>>();
    }
    public EMMMapBase(EMMMapBase<T> original, Cloner cloner) {
      if (original.ModelSet != null) {
        if (original.ModelSet is List<ISymbolicExpressionTree> originalSet && ModelSet is List<ISymbolicExpressionTree> set)
          set = originalSet.Select(cloner.Clone).ToList();
        else ModelSet = original.ModelSet.ToList(); /// check this if you want to use it
      }
      if (original.ClusterNumber != null) {
        ClusterNumber = original.ClusterNumber.ToList();
      }
      if (original.Map != null) {
        Map = original.Map.Select(x => x.ToList()).ToList();
      }
      K = original.K;
    }
    //public EMMMapBase(IRandom random, IEnumerable<T> trees, int k) : this(k) {
    //  // constructor that should be used in case of creation of a new map from the start point
    //  ModelSet = trees.ToList();
    //  CalculateDistances();
    //  CreateMap(random, k);
    //}
    //public EMMMapBase(IRandom random, IEnumerable<T> trees, string fileName = "Map.txt") : this(1) {
    //  // constructor that shoud be used in case of using of "old" map, that was previously created and now shoud be readed from file
    //  ModelSet = trees.ToList();
    //  MapFromFileRead(fileName);
    //  K = Map.Count;
    //  MapPreparation();
    //}
    public void MapCreationPrepare(IRandom random, IEnumerable<T> trees, int k) {
      ModelSet = trees.ToList();
      CalculateDistances();
      CreateMap(random, k);
    }
    public void MapRead(IRandom random, IEnumerable<T> trees, string fileName = "Map.txt") {
      ModelSet = trees.ToList();
      MapFromFileRead(fileName);
      K = Map.Count;
      MapPreparation();
      if (this is EMMNetworkMap one) { one.NeghboorNumber = Map[0].Count; }
    }
    public void WriteMapToTxtFile(IRandom random) {
      string s = random.ToString();
      string fileName = "Map";
      fileName += s;
      fileName += ".txt";
      File.WriteAllLines(fileName, MapToString());
      string fileName2 = "MapToSee";
      fileName2 += s;
      fileName2 += ".txt";
      File.WriteAllLines(fileName, MapToSee());
    }
    public string[] MapToString() { // Function that preapre Map to printing in .txt File: create a set of strings for future reading by computer
      string[] s;
      s = new string[K];
      for (int i = 0; i < K; i++) {
        s[i] = "";
        for (int j = 0; j < Map[i].Count; j++) {
          s[i] += Map[i][j].ToString();
          s[i] += " ";
        }
      }
      return s;
    }
    public string[] MapToSee() { // Function that prepare Map to printing in .txt File: create a set of strings in human readable view
      var fmt = new InfixExpressionFormatter();
      string[] s;
      s = new string[(ModelSet.Count) + 1];
      s[0] = "ClusterNumber" + "," + "ModelNumber" + "," + "Model";
      for (int i = 1; i < ((ModelSet.Count) + 1); i++) {
        s[i] = ClusterNumber[i - 1].ToString() + "," + (i - 1).ToString() + ",";
        if (ModelSet[i - 1] is ISymbolicExpressionTree model) {
          s[i] += fmt.Format(model);
        } else { s[i] += ModelSet[i - 1].ToString(); }
      }
      return s;
    }
    public void MapFromFileRead(string fileName) {
      string input = File.ReadAllText(fileName);
      int i = 0;
      foreach (var row in input.Split('\n')) {
        Map.Add(new List<int>());
        foreach (var col in row.Trim().Split(' ')) {
          Map[i].Add(int.Parse(col.Trim()));
        }
        i++;
      }
    }
    protected void MapSizeCheck() {
      if (Map != null) Map.Clear();
      else Map = new List<List<int>>();
      if (Map.Count != K) {
        if (Map.Count != 0) {
          Map.Clear();
        }
        for (int i = 0; i < K; i++) {
          Map.Add(new List<int>());
        }
      }
    }
    protected void MapPreparation() {
      for (int i = 0; i < Map.Count; i++) {
        for (int j = 0; j < Map[i].Count; j++) {
          ClusterNumber.Add(0);
        }
      }
      for (int i = 0; i < Map.Count; i++) {
        for (int j = 0; j < Map[i].Count; j++) {
          ClusterNumber[Map[i][j]] = i;
        }
      }
    }
  }
}
